unit u_mnk;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, u_data;

var
 MNKp: Cardinal;
 a0,a1,a2: double;
 mnkX, mnkY: array [0..MAX_POINTS] of single;
 func: byte;

procedure calc_mnk_lin; //y(x)=a1*x + a0
procedure calc_mnk_par; //y(x)=a2*x^2 + a1*x + a0
procedure calc_mnk_exp;  //y(x)=a0 * exp (a1*x)
procedure calc_mnk_ln;  //y(x)=a1*ln(x) + a0
//Расчёт всех функций по МНК и выбор лучшей
procedure calc_mnk; //1 - lin, 2 - par, 3 - exp, 4 - ln
//Лучшая функция МНК после calc_mnk
function func_mnk(x: single): single;

implementation

procedure calc_mnk;
var
 n: integer;
 K: array [1..4, 0..2] of double;
 sko: array [1..4] of double;
begin
 func:=0;
 if (MNKp < 1) then exit;
 if (MNKp < 3) then
  begin
   func:=1;
   calc_mnk_lin;
   exit;
  end;
 mnkX[MNKp]:=0; // Добавим одну нулевую точку
 mnkY[MNKp]:=0; // для избежания получения неправильной функции
 inc(MNKp);     // при недостатке данных
 calc_mnk_lin; K[1,0]:=a0; K[1,1]:=a1; K[1,2]:=a2;
 calc_mnk_par; K[2,0]:=a0; K[2,1]:=a1; K[2,2]:=a2;
 calc_mnk_exp; K[3,0]:=a0; K[3,1]:=a1; K[3,2]:=a2;
 calc_mnk_ln; K[4,0]:=a0; K[4,1]:=a1; K[4,2]:=a2;
 for n:=1 to 4 do sko[n]:=0;
 for n:=0 to MNKp-1 do sko[1]:=sko[1]+sqr(mnkY[n] - K[1,1]*mnkX[n] - K[1,0]);
 for n:=0 to MNKp-1 do
  sko[2]:=sko[2]+sqr(K[2,2]*sqr(mnkX[n]) + K[2,1]*mnkX[n] + K[2,0] - mnkY[n]);
 for n:=0 to MNKp-1 do sko[3]:=sko[3]+sqr(K[3,0]*exp(K[3,1]*mnkX[n]) - mnkY[n]);
 for n:=0 to MNKp-1 do if (mnkX[n]>0) then
  sko[4]:=sko[4]+sqr(K[4,1]*ln(mnkX[n]) + K[4,0] -mnkY[n]);
 func:=1;
 for n:=2 to 4 do
  if (sko[func] > sko[n]) then func:=n;
 a0:=K[func,0]; a1:=K[func,1]; a2:=K[func,2];
end;

function func_mnk(x: single): single;
begin
 case func of
  1: func_mnk:=a1*x + a0;
  2: func_mnk:=a2*sqr(x) + a1*x + a0;
  3: func_mnk:=a0 * exp(a1*x);
  4: if (x>0) then func_mnk:=a1*ln(x) + a0 else func_mnk:=a0;
  else func_mnk:=0;
 end;
end;

procedure calc_mnk_lin;
var
 n: integer;
 Sx, Sy, Sx2, Sxy: double;
begin
 a0:=0; a1:=0; a2:=0;
 if (MNKp<1) then exit;
 if (MNKp=1) then
  begin
   if (mnkX[0]<>0) then a1:=mnkY[0]/mnkX[0] else a1:=0;
   exit;
  end;
 Sx:=0; Sy:=0; Sx2:=0; Sxy:=0;
 for n:=0 to (MNKp-1) do
  begin
   Sx:=Sx+mnkX[n];
   Sy:=Sy+mnkY[n];
   Sx2:=Sx2+mnkX[n]*mnkX[n];
   Sxy:=Sxy+mnkX[n]*mnkY[n];
  end;
 a0:=(Sy - Sx*Sxy/Sx2) / (MNKp - Sx*Sx/Sx2);
 a1:=(Sxy - Sx*a0) / Sx2;
end;

procedure calc_mnk_par;
var
 My, Myx, Myx2, Mx, Mx2, Mx3, Mx4, X0, X1, X2: double;
 i,j,k: integer;
 mat: array [1..3, 1..4] of double;
 buf: array [1..4] of double;
begin
 a0:=0; a1:=0; a2:=0;
 if (MNKp<1) then exit;
 if (MNKp=1) then begin a1:=mnkY[0]/mnkX[0]; exit; end;
 if (MNKp=2) then
  begin
   if (mnkX[1]=mnkX[0]) then begin a1:=mnkX[0]/mnkY[0]; exit; end;
   a1:=abs(mnkY[1]-mnkY[0])/abs(mnkX[1]-mnkX[0]);
   a0:=abs(mnkX[1]*mnkY[0]-mnkY[1]*mnkX[0])/abs(mnkX[1]-mnkX[0]);
   exit;
  end;
 Mx:=0; Mx2:=0; Mx3:=0; Mx4:=0; My:=0; Myx:=0; Myx2:=0;
 for i:=0 to MNKp-1 do
  begin
   Mx:= Mx + mnkX[i];
   Mx2:= Mx2 + mnkX[i]*mnkX[i];
   Mx3:= Mx3 + mnkX[i]*mnkX[i]*mnkX[i];
   Mx4:= Mx4 + mnkX[i]*mnkX[i]*mnkX[i]*mnkX[i];
   My:= My + mnkY[i];
   Myx:= Myx + mnkY[i] * mnkX[i];
   Myx2:= Myx2 + mnkY[i] * mnkX[i]*mnkX[i];
  end;
 mat[1,1]:= Mx2; mat[1,2]:= Mx; mat[1,3]:= MNKp; mat[1,4]:= My;
 mat[2,1]:= Mx3; mat[2,2]:= Mx2; mat[2,3]:= Mx; mat[2,4]:= Myx;
 mat[3,1]:= Mx4; mat[3,2]:= Mx3; mat[3,3]:= Mx2; mat[3,4]:= Myx2;
 i:= 1;
 while (mat[1,1] = 0) or (i > 3) do
  begin
   for j:= 1 to 4 do
    begin
     buf[j]:= mat[1,j];
     mat[1,j]:= mat[i+1,j];
     mat[i+1,j]:= buf[j];
    end;
   i:= i + 1;
  end;
 for i:= 4 downto 1 do
  mat[1,i]:= mat[1,i] / mat[1,1];
 k:= 2;
 for i:=1 to 2 do
  begin
   for j:= 4 downto 1 do
    mat[k,j]:= mat[k,j] - mat[1,j] * mat[k,1];
   k:= k + 1;
  end;
 for i:= 4 downto 1 do
  mat[2,i]:= mat[2,i] / mat[2,2];
 for j:= 4 downto 2 do
  mat[3,j]:= mat[3,j] - mat[2,j] * mat[3,2];
 a0:= mat[3,4] / mat[3,3];
 a1:= mat[2,4] - mat[2,3] * a0;
 a2:= mat[1,4] - mat[1,3] * a0 - mat[1,2] * a1;
end;

procedure calc_mnk_exp;
var
 i: integer;
 s1,s2,s3,s4,D,x1,y1:double;
begin
 s1:=0; s2:=0; s3:=0; s4:=0; a0:=0; a1:=0; a2:=0;
 if (MNKp<1) then exit;
 For i:=1 to MNKp do
  begin
   if (mnkY[i]<=0) then continue;
   D:=ln(mnkY[i]);
   s1:=s1+mnkX[i]; s2:=s2+D;
   s3:=s3+mnkX[i]*D; s4:=s4+mnkX[i]*mnkX[i]
  end;
 D:=MNKp*s4-sqr(s1);
 if (D=0) then exit;
 a1:=(MNKp*s3-s1*s2)/D; a0:=exp((s4*s2-s1*s3)/D);
end;

procedure calc_mnk_ln;
var
 n: integer;
 Sx, Sy, Sx2, Sxy, lnx: double;
begin
 a0:=0; a1:=0; a2:=0;
 if (MNKp<1) then exit;
 if (MNKp=1) then begin a1:=mnkY[0]/mnkX[0]; exit; end;
 Sx:=0; Sy:=0; Sx2:=0; Sxy:=0;
 for n:=0 to (MNKp-1) do
  begin
   if (mnkX[n] > 0) then lnx:=ln(mnkX[n]) else lnx:=0;
   Sx:=Sx+lnx;
   Sy:=Sy+mnkY[n];
   Sx2:=Sx2+lnx*lnx;
   Sxy:=Sxy+lnx*mnkY[n];
  end;
 a0:=(Sy - Sx*Sxy/Sx2) / (MNKp - Sx*Sx/Sx2);
 a1:=(Sxy - Sx*a0) / Sx2;
end;

end.

